// Client.cc
//
// Purpose:
// 	Handles creation, life and death of a client window.
// 	Handles creation of window decorations, resizing, shading, etc...
// 	Handles all events associated with a client window.
//
// 
//
//
// Frank Hale
// frankhale@yahoo.com / frankhale@mailcity.com
// 17 Jan 2000
/////////////////////////////////////////////////////////////////////////

#include "sapphire.hh"

Client::Client() 
{
	// This takes care of assigning default values for the client object.

	// Shaded means only the titlebar is showing.		
	shaded=False;
	
	// Maxsized means if the client is maximized.
	maxsized=False;
	
	// Since we don't know the name of the client we will just assign it a default name.
	name="unknown";
	
	// Defaults for this Clients x, y position and width and height. 
	x=1;
	y=1;
	width=1;
	height=1;
	
	oumx=1;
	oumy=1;
	oumwidth=1;
	oumheight=1;

	default_titlebar_height = wm->iheight + DEFAULT_SPACE;

	default_client_name_x = DEFAULT_SPACE+ 25;
	default_client_name_y = DEFAULT_SPACE + wm->font->ascent + 2;

	default_button_width=15;
	default_button_height=15;

	position_window_width=1;
	position_window_height=1;

	// These are set to there real defaults in the reparent function.
	default_close_button_x = 0;
	default_close_button_y = 0;
	default_shade_button_x = 0;
	default_shade_button_y = 0;
	default_minimize_button_x = 0;
	default_minimize_button_y = 0;	
	default_maximize_button_x = 0;
	default_maximize_button_y = 0;

	close_button_pressed=False;
	shade_button_pressed=False;
	minimize_button_pressed=False;
	maximize_button_pressed=False;

	// This client starts off not having the focus. Focus follows mouse pointer.
	has_focus=False;
	
	// Mouse Button defaults
	button1=button2=button3=False;
	
	// Window Decorations
	//
	// The parent_window is the main parent. The titlebar is a child of the parent_window
	// all other button windows are childs of the titlebar.
	has_titlebar=True;
	has_close_button=True;
	has_minimize_button=True;
	has_maximize_button=True;
	has_shade_button=True;

	close.window = maximize.window 
		     = minimize.window 
		     = shadeButton.window
		     = window 
		     = parent 
		     = transient 
		     = title
		     = win_pos 
		     = None;
}

Client::~Client() 
{
	// Nothing to do here
}

// function: Window Client::createTopLevelWindow(int x, int y, 
//	unsigned int width,unsigned int height,unsigned int borderwidth)
//
// Purpose:
//
// Handles creation of a top level window. Top level windows are children
// of the root window. 
//
// Returns:
//
// This function returns the top level window it created.
//
Window Client::createTopLevelWindow(int x, int y, 
	unsigned int width,unsigned int height,unsigned int borderwidth)
{
	XSetWindowAttributes attrib;

	unsigned long create_mask = CWOverrideRedirect | 
		CWBackPixmap | CWBackPixel | CWBorderPixel | CWCursor | 
		CWEventMask;

	attrib.background_pixmap = None;
	attrib.background_pixel = wm->background_color.pixel;
	attrib.border_pixel = wm->border_color.pixel;
	attrib.override_redirect = True;
	attrib.cursor = XCreateFontCursor(wm->getDisplay(), XC_left_ptr);
	attrib.event_mask =  StructureNotifyMask | SubstructureNotifyMask |
		SubstructureRedirectMask | ExposureMask | ButtonMotionMask | 
		ButtonPressMask | ButtonReleaseMask | FocusChangeMask;
	
	return (XCreateWindow(wm->getDisplay(), wm->getRootWindow(), x,y,width,height,borderwidth,
		DefaultDepth(wm->getDisplay(), wm->screen), InputOutput, DefaultVisual(wm->getDisplay(), wm->screen),
		create_mask, &attrib));
}

// function: Window Client::createChildWindow(Window parent, int x, int y, 
//	unsigned int width,unsigned int height,unsigned int borderwidth)
//
// Purpose:
//
// Handles creation of a child window. You specify the window that is its
// parent.
//
// Returns:
//
// This function returns the child window that it created.
//
Window Client::createChildWindow(Window parent, int x, int y, 
	unsigned int width,unsigned int height,unsigned int borderwidth)
{
	XSetWindowAttributes attrib;

	unsigned long create_mask = 
		CWBackPixmap | CWBackPixel | CWBorderPixel | CWCursor | 
		CWEventMask;

	attrib.background_pixmap = None;
	attrib.background_pixel = wm->background_color.pixel;
	attrib.border_pixel = wm->border_color.pixel;
	attrib.cursor = XCreateFontCursor(wm->getDisplay(), XC_left_ptr);
	attrib.event_mask = StructureNotifyMask | SubstructureNotifyMask |
		SubstructureRedirectMask | ExposureMask | EnterWindowMask |
		LeaveWindowMask | ButtonMotionMask | ButtonPressMask | 
		ButtonReleaseMask;
	
	return (XCreateWindow(wm->getDisplay(), parent, x,y,width,height,borderwidth,
		DefaultDepth(wm->getDisplay(), wm->screen), InputOutput, DefaultVisual(wm->getDisplay(), wm->screen),
		create_mask, &attrib));
}

// function: void Client::make_new_client(Window w)
//
// Purpose:
//
// Handles creation of a client window. It positions, calls reparent to create window decorations
// and inserts the newly created client into the client list.
//
// Returns:
// 
// Nothing
//
void Client::make_new_client(Window w)
{
	/* For WM_PROTOCOLS handling. */
	Atom * protocols;
	int n;
	int p;

	long dummy;
	XWMHints *hints;

	#ifdef DEBUG
		printf("started Client::make_new_client(Window w)\n");
	#endif

	// Populate the hints structure with the window hints (w) has.
	hints = XGetWMHints(wm->getDisplay(), w);

	// Create a new Client object.
	Client *c = new Client();

	// Insert the client into the window manager client list.
	wm->insertClient(c);

	// Assign the window (w) to c->window, c->window represents the application.
	c->window = w;
	c->ignore_unmap = 0;

	XSelectInput(wm->getDisplay(), c->window, ColormapChangeMask | PropertyChangeMask | EnterWindowMask );

	/*
	 * Scan the list of atoms on WM_PROTOCOLS to see which of the
	 * protocols that we understand the client is prepared to
	 * participate in. (See ICCCM section 4.1.2.7.)
	 */
	if (XGetWMProtocols(wm->getDisplay(), c->window, &protocols, &n) != 0) {
		for (p = 0; p < n; p++) {
			if (protocols[p] == wm->getWMDeleteAtom()) {
				c->proto |= Pdelete;
			}
			else if (protocols[p] == wm->getWMFocusAtom())
				c->proto |= Ptakefocus;
		}

		XFree(protocols);
	}
	
	getTransients(c);

	XFetchName(wm->getDisplay(), c->window, &c->name);

	c->size = XAllocSizeHints();
	XGetWMNormalHints(wm->getDisplay(), c->window, c->size, &dummy);

	if (hints->flags & StateHint) {
		set_wm_state(c, hints->initial_state);
	}
	
	if (hints) XFree(hints);
	
	if (isViewable(c)) c->ignore_unmap++;

	change_gravity(c, 1);

	// Determine if this client needs window decorations.
	setButtonPositionDefaults(c);
	determineWindowDecorations(c);
	
	// Do all the necessary reparenting for this client
	reparent(c);
	
	// If this client should be iconic on start then we won't map it.
	isIconic(c);
	
	XAddToSaveSet(wm->getDisplay(), c->window);		

#ifdef DEBUG
	// This prints off some useful information about the newly created client.
	err("%s (%d, %d): %#x (%#x) @ %d,%d", c->name, get_wm_state(c),
		c->ignore_unmap, c->window, c->parent, c->x, c->y);
#endif

	#ifdef DEBUG
		printf("finished Client::make_new_client(Window w)\n");
	#endif	
}

// Function: void Client::remove_client(Client *c, int from_cleanup)
//
// Purpose: 
//
// Handles removing a client from client list. It destroys all window decorations and free's pixmaps. 
// It handles all clean up associated with a client window.
//
void Client::remove_client(Client *c)
{
	XEvent foo;
	
	#ifdef DEBUG
		printf("started Client::remove_client(Client *c)\n");
	#endif

	XGrabServer(wm->getDisplay());
	
	if(! XCheckTypedWindowEvent(wm->getDisplay(), c->window, DestroyNotify, &foo))
	{
		// Sets input focus back to the root window
		XSetInputFocus(wm->getDisplay(), wm->getRootWindow(), RevertToPointerRoot, CurrentTime);
		
		// Get the windows state if it was Iconic then map it. 
		if (get_wm_state(c) == IconicState) XMapWindow(wm->getDisplay(), c->window);

		change_gravity(c, -1);
		
		XDestroyWindow(wm->getDisplay(), c->close.window);
		XDestroyWindow(wm->getDisplay(), c->maximize.window);
		XDestroyWindow(wm->getDisplay(), c->minimize.window);
		XDestroyWindow(wm->getDisplay(), c->shadeButton.window);
		XDestroyWindow(wm->getDisplay(), c->title);
		XDestroyWindow(wm->getDisplay(), c->win_pos);
		XDestroyWindow(wm->getDisplay(), c->parent);
					
		if (c->name) XFree(c->name);
		if (c->size) XFree(c->size);

		XFreePixmap(wm->getDisplay(), c->frame_pix);
		XFreePixmap(wm->getDisplay(), c->button_pix);
		XFreePixmap(wm->getDisplay(), c->focus_pix);
		XFreePixmap(wm->getDisplay(), c->position_pix);
		
		wm->removeClient(c);
	
		delete c;
	} 
	
	XSync(wm->getDisplay(), False);	
	XUngrabServer(wm->getDisplay());
		
	#ifdef DEBUG
		printf("finished Client::remove_client(Client *c, int from_cleanup)\n");
	#endif	
}

// Function:
//
// 	void Client::setButtonPositionDefaults(Client *c)
//
// Description:
//
// 	The default positions for the titlebar buttons are as follows
//
// 	[close_button] Client Name [shade_button][minimize_button][maximize_button]
//
// 	There associated x,y pairs are:
//
// 	c->default_close_button_x = 4;
// 	c->default_close_button_y = 4;
//
// 	c->default_shade_button_x = c->width - c->default_titlebar_height - 36;
// 	c->default_shade_button_y = 4;
//
// 	c->default_minimize_button_x = c->width - c->default_titlebar_height - 18;
// 	c->default_minimize_button_y = 4;	
//
// 	c->default_maximize_button_x = c->width - c->default_titlebar_height;
// 	c->default_maximize_button_y = 4;
//
// Returns:
// 
// 	Nothing
void Client::setButtonPositionDefaults(Client *c)
{
	c->default_close_button_x = 4;
	c->default_close_button_y = 4;
	
	c->default_shade_button_x = c->width - c->default_titlebar_height - 36;
	c->default_shade_button_y = 4;
	
	c->default_minimize_button_x = c->width - c->default_titlebar_height - 18;
	c->default_minimize_button_y = 4;	
	
	c->default_maximize_button_x = c->width - c->default_titlebar_height;
	c->default_maximize_button_y = 4;
}

void Client::reparent(Client *c)
{		
	#ifdef DEBUG
		printf("started Client::reparent(Client *c)\n");
	#endif

	// Since very few applications ever request that they not be reparented we 
	// will still create the window decorations regardless. In the event that a 
	// particular client doesn't wish to have window decorations we simply won't 
	// show them. This may be a wrong to do but it works and is of low priority
	// at this point in time to change it.
		
	// In order to create window decorations and buttons for resizing,
	// shading, etc.. we need to create a few extra windows.
	createParentWindow(c);
	createTitlebarWindow(c);
	createCloseButtonWindow(c);
	createMinimizeButtonWindow(c);
	createMaximizeButtonWindow(c);
	createShadeButtonWindow(c);
	createPositionWindow(c);

//#ifdef SHAPE
//	if (wm->shape) {
//		XShapeSelectInput(wm->getDisplay(), c->window, ShapeNotifyMask);
//		set_shape(c);
//	}
//#endif
	
	// If this window doesn't request that it shouldn't be reparented
	// then we will go ahead an reparent it to c->parent which is the 
	// parent window which holds the client application window ( c->window ) aswell
	// as the titlebar window and the button windows.
	if (c->has_titlebar) XReparentWindow(wm->getDisplay(), c->window, c->parent, 0, theight(c));
	else 
	{
		// If we don't need to reparent it with window decorations then we
		// reparent it back to the root window.
		XReparentWindow(wm->getDisplay(), c->window, wm->getRootWindow(), 0, 0);
	}

	reconfigure(c);		
	send_config(c);

	#ifdef DEBUG
		printf("finished Client::reparent(Client *c)\n");
	#endif
}

void Client::createParentWindow(Client *c)
{
	c->parent = createTopLevelWindow(c->x,c->y - theight(c),
			c->width, c->height + theight(c), DEFAULT_BORDER_SIZE);
}

void Client::createTitlebarWindow(Client *c)
{
	c->title = createChildWindow(c->parent, 0, 0, c->width, theight(c),0);
}

void Client::createCloseButtonWindow(Client *c)
{
	c->close.x      = c->default_close_button_x;
	c->close.y      = c->default_close_button_y;
	c->close.width  = c->default_button_width;
	c->close.height = c->default_button_height;
	
	c->close.window = 
		createChildWindow(c->title, c->close.x, c->close.y
		, c->close.width, c->close.height,
		1);
}

void Client::createMinimizeButtonWindow(Client *c)
{
	c->minimize.x = c->default_minimize_button_x;
	c->minimize.y = c->default_minimize_button_y;
	c->minimize.width =  c->default_button_width;
	c->minimize.height =  c->default_button_height;

	c->minimize.window = 
	createChildWindow(c->title, c->minimize.x, c->minimize.y
	, c->minimize.width, c->minimize.height,
	1);
}

void Client::createMaximizeButtonWindow(Client *c)
{
	c->maximize.x      = c->default_maximize_button_x;
	c->maximize.y      = c->default_maximize_button_y;
	c->maximize.width  = c->default_button_width;
	c->maximize.height = c->default_button_height;

	c->maximize.window = 
		createChildWindow(c->title, c->maximize.x, c->maximize.y
		, c->maximize.width, c->maximize.height,
		1);
}

void Client::createShadeButtonWindow(Client *c)
{
	c->shadeButton.x      = c->default_shade_button_x;
	c->shadeButton.y      = c->default_shade_button_y;
	c->shadeButton.width  = c->default_button_height;
	c->shadeButton.height = c->default_button_height;

	c->shadeButton.window = 
		createChildWindow(c->title, c->shadeButton.x, c->shadeButton.y
		, c->shadeButton.width, c->shadeButton.height,
		1);	
}

void Client::createPositionWindow(Client *c)
{
	int x=wm->getXRes()+100; // Make sure this thing is off screen somewhere
	int y=wm->getYRes()+100; //
	int width=1;
	int height=1;

	c->win_pos = createTopLevelWindow(
		x,		 		/* x position */
		y, 				/* y position */
		width, 				/* width */
		height, 			/* height */
		1				/* Border width */
		);
}

int Client::theight(Client *c) const 
{ 
	//return ((c->transient ? 2*DEFAULT_SPACE : wm->iheight) + DEFAULT_SPACE); 

	// For now this function will pass a Client object. Even though we don't 
	// use it for much, I still might want to have transients have a different style
	// than there top level parents later on.
	//
	// For instance I could turn off some of the titlebar buttons, or I could change the
	// height of the client's titlebar and button width and height. 
	//
	// This is the reason why I am passing the Client object to this function. I simply
	// haven't made up my mind at this point.
	return c->default_titlebar_height; 
} 

void Client::getTransients(Client *c)
{
	Window trans;
	
	XGetTransientForHint(wm->getDisplay(), c->window, &trans);
	
	c->transient = trans;
}

void Client::setCloseButtonPosition(Client *c, int x, int y)
{
	c->default_close_button_x = c->close.x = x;
	c->default_close_button_y = c->close.y = y;
}

void Client::setShadeButtonPosition(Client *c, int x, int y)
{
	c->default_shade_button_x = c->shadeButton.x = x;
	c->default_shade_button_y = c->shadeButton.y = y;
}

void Client::setMinimizeButtonPosition(Client *c, int x, int y)
{
	c->default_minimize_button_x = c->minimize.x = x;
	c->default_minimize_button_y = c->minimize.y = y;
}

void Client::setMaximizeButtonPosition(Client *c, int x, int y)
{
	c->default_maximize_button_x = c->maximize.x = x;
	c->default_maximize_button_y = c->maximize.y = y;
}

void Client::setClientNamePosition(Client *c, int x, int y)
{
	c->default_client_name_x = x;
	c->default_client_name_y = y;
}

xy_pair Client::getCloseButtonPosition(Client *c) const
{
	xy_pair xy;
	
	xy.x = c->default_close_button_x;
	xy.y = c->default_close_button_y;
	
	return xy;
}

xy_pair Client::getShadeButtonPosition(Client *c) const
{
	xy_pair xy;
	
	xy.x = c->default_shade_button_x;
	xy.y = c->default_shade_button_y;
	
	return xy;
}

xy_pair Client::getMinimizeButtonPosition(Client *c) const
{
	xy_pair xy;
	
	xy.x = c->default_minimize_button_x;
	xy.y = c->default_minimize_button_y;
	
	return xy;
}

xy_pair Client::getMaximizeButtonPosition(Client *c) const
{
	xy_pair xy;
	
	xy.x = c->default_maximize_button_x;
	xy.y = c->default_maximize_button_y;
	
	return xy;
}

xy_pair Client::getClientNamePosition(Client *c) const
{
	xy_pair xy;
	
	xy.x = c->default_client_name_x;
	xy.y = c->default_client_name_y;
	
	return xy;
}

void Client::determineWindowDecorations(Client *c)
{
	// check for mwm hints and change the window decorations if necessary
	//
	// But since I am only worried about whether a client needs a titlebar or
	// not these will probably never be true compliant with Motif. 
	//
	// I personally like my clients to have all the buttons a top level window has.
	// If it doesn't need a titlebar then I can live with that and we won't map 
	// one.
	// 
	// For example: xmms looks like shit with a titlebar mapped to it. This is mainly
	// the whole reason behind this function so apps like xmms look correctly.
	//
	int format;
	Atom atom_return;
	unsigned long num, len;
      
	if (XGetWindowProperty(wm->getDisplay(), c->window,
                         wm->getMotifWMHintsAtom(), 0,
                         PropMwmHintsElements, False,
                         wm->getMotifWMHintsAtom(), &atom_return,
                         &format, &num, &len,
                         (unsigned char **) &c->mwm_hint) == Success && c->mwm_hint)
	
	if (num == PropMwmHintsElements) {
		if (c->mwm_hint->flags & MwmHintsDecorations)
		{
			if (c->mwm_hint->decorations & MwmDecorAll) 
			{
				c->has_titlebar=c->has_close_button=c->has_minimize_button=c->has_maximize_button=c->has_shade_button=True;
	    		}
	    		else {
				c->has_titlebar=c->has_close_button=c->has_minimize_button=c->has_maximize_button=c->has_shade_button=False;
	      
	      			if (c->mwm_hint->decorations & MwmDecorTitle) c->has_titlebar = True;
				
				//if (c->mwm_hint->decorations & MwmDecorIconify) c->has_minimize_button = True;
	    
	      			//if (c->mwm_hint->decorations & MwmDecorMaximize) c->has_maximize_button = True;
		
				// Well I like all my windows to have a shade button and a close button
				
				c->has_minimize_button = True;
				c->has_maximize_button = True;
			
				c->has_shade_button=True;
				c->has_close_button=True;
				
				//if(! c->has_minimize_button)
				//{
				//	setShadeButtonPosition(c, c->default_minimize_button_x, c->default_minimize_button_y);	
				//} 
				
				//if (! c->has_maximize_button)
				//{
				//	setShadeButtonPosition(c, c->default_maximize_button_x, c->default_maximize_button_y);
				//} 				  
	    		}
		}
	}
}

// function: Client *Client::find_client(Window w)
//
// Purpose:
//
// This function will look through the client list for a window which matches the one
// passed to it. If it finds it, it will return the client object. 
//
Client *Client::find_client(Window w)
{
	if (w && w != wm->getRootWindow()) {
		if(wm->getClientCount()) {
			LinkedListIterator<Client> it(wm->getClientListObject());
			for(; it.current(); it++)
			{
				if (w == it.current()->parent ||
				    w == it.current()->title ||
				    w == it.current()->window ||
				    w == it.current()->close.window ||
				    w == it.current()->maximize.window ||
				    w == it.current()->minimize.window ||
				    w == it.current()->shadeButton.window )
				{
					return it.current();
				}				
			}
		}
	}

	return NULL;
}

int Client::isViewable(Client *c)
{
	XWindowAttributes attr;

	XGetWindowAttributes(wm->getDisplay(), c->window, &attr);
		
	c->x = attr.x;
	c->y = attr.y;
	c->width = attr.width;
	c->height = attr.height;
	c->cmap = attr.colormap;
	c->override_redirect = attr.override_redirect;
	
	return attr.map_state;
}

void Client::createTitleImage(Client* c)
{
	if(c->Title.getWidth() != c->width)
	{
		c->Title.setWidth(c->width);
		c->Title.setHeight(theight(c));		
		c->Title.setColorDepth(DefaultDepth(wm->getDisplay(), wm->screen));
		c->Title.createImage();

		if(wm->theme->isTitlebarHorizontalGradient())
			c->Title.renderHGradient(wm->theme->getClientColorFrom(), wm->theme->getClientColorTo());
		
		if(wm->theme->isTitlebarVerticalGradient())
			c->Title.renderVGradient(wm->theme->getClientColorFrom(), wm->theme->getClientColorTo());
	
		if(wm->theme->isTitlebarSolid())
			c->Title.setBackgroundColor(wm->theme->getClientColor());
	
		if(wm->theme->isTitlebarBeveled())
			c->Title.renderBevel();
	
		if (c->frame_pix) XFreePixmap(wm->getDisplay(), c->frame_pix);
		
		c->frame_pix = c->Title.convertToPixmap(wm->getDisplay(), DefaultVisual(wm->getDisplay(), wm->screen), c->title);
	}
}

void Client::createButtonImage(Client* c)
{
	if(c->Button.getWidth() != 15)
	{
		c->Button.setWidth(15);
		c->Button.setHeight(theight(c));
		c->Button.setColorDepth(DefaultDepth(wm->getDisplay(), wm->screen));
		c->Button.createImage();
	
		if(wm->theme->isButtonHorizontalGradient())
			c->Button.renderHGradient(wm->theme->getTitleBarButtonColorFrom(),wm->theme->getTitleBarButtonColorTo());

		if(wm->theme->isButtonVerticalGradient())
			c->Button.renderVGradient(wm->theme->getTitleBarButtonColorFrom(),wm->theme->getTitleBarButtonColorTo());

		if(wm->theme->isButtonSolid())
			c->Button.setBackgroundColor(wm->theme->getTitleBarButtonColor());
		
		if(wm->theme->isButtonBeveled())
			c->Button.renderBevel();
	
		if (c->button_pix) XFreePixmap(wm->getDisplay(), c->button_pix);
	
		c->button_pix = c->Button.convertToPixmap(wm->getDisplay(), DefaultVisual(wm->getDisplay(), wm->screen), c->shadeButton.window);	

		XSetWindowBackgroundPixmap(wm->getDisplay(), c->shadeButton.window, c->button_pix);
		XSetWindowBackgroundPixmap(wm->getDisplay(), c->close.window, c->button_pix);
		XSetWindowBackgroundPixmap(wm->getDisplay(), c->minimize.window, c->button_pix);
		XSetWindowBackgroundPixmap(wm->getDisplay(), c->maximize.window, c->button_pix);

	}
}

void Client::createFocusImage(Client* c)
{
	if(c->Focus.getWidth() != c->width)
	{	
		c->Focus.setWidth(c->width);
		c->Focus.setHeight(theight(c)+1);
		c->Focus.setColorDepth(DefaultDepth(wm->getDisplay(), wm->screen));
		c->Focus.createImage();
	
		if(wm->theme->isTitlebarHorizontalGradient())
			c->Focus.renderHGradient(wm->theme->getClientFocusColorFrom(), wm->theme->getClientFocusColorTo());

		if(wm->theme->isTitlebarVerticalGradient())
			c->Focus.renderVGradient(wm->theme->getClientFocusColorFrom(), wm->theme->getClientFocusColorTo());
	
		if(wm->theme->isTitlebarSolid())
			c->Focus.setBackgroundColor(wm->theme->getClientFocusColor());
		
		if(wm->theme->isTitlebarBeveled())
			c->Focus.renderBevel();
	
		if (c->focus_pix) XFreePixmap(wm->getDisplay(), c->focus_pix);
	
		c->focus_pix = c->Focus.convertToPixmap(wm->getDisplay(), DefaultVisual(wm->getDisplay(), wm->screen), c->title);		
	}	
}

void Client::createPositionImage(Client* c)
{
	getClientPosition(c);

	if(c->Position.getWidth() != c->position_window_width)
	{	
		c->Position.setWidth(c->position_window_width);
		c->Position.setHeight(c->position_window_height);
		c->Position.setColorDepth(DefaultDepth(wm->getDisplay(), wm->screen));
		c->Position.createImage();
	
		if(wm->theme->isPositionHorizontalGradient())
			c->Position.renderHGradient(wm->theme->getPositionColorFrom(), wm->theme->getPositionColorTo());

		if(wm->theme->isPositionVerticalGradient())
			c->Position.renderVGradient(wm->theme->getPositionColorFrom(), wm->theme->getPositionColorTo());
	
		if(wm->theme->isPositionSolid())
			c->Position.setBackgroundColor(wm->theme->getPositionColor());
		
		if(wm->theme->isPositionBeveled())
			c->Position.renderBevel();
	
		if (c->position_pix) XFreePixmap(wm->getDisplay(), c->position_pix);
	
		c->position_pix = c->Position.convertToPixmap(wm->getDisplay(), DefaultVisual(wm->getDisplay(), wm->screen), c->win_pos);		
	
		XSetWindowBackgroundPixmap(wm->getDisplay(), c->win_pos, c->position_pix);
	}
}

void Client::do_gradient(Client *c)
{      		
	if (c->has_titlebar)
	{		
		XClearWindow(wm->getDisplay(), c->title);
		
		if(! c->has_focus)
		{		
			XSetWindowBackgroundPixmap(wm->getDisplay(), c->title, c->focus_pix);	
		}
		else XSetWindowBackgroundPixmap(wm->getDisplay(), c->title, c->frame_pix);

				
		XClearWindow(wm->getDisplay(), c->close.window);
		XClearWindow(wm->getDisplay(), c->minimize.window);		
		XClearWindow(wm->getDisplay(), c->maximize.window);
		XClearWindow(wm->getDisplay(), c->shadeButton.window);
		XClearWindow(wm->getDisplay(), c->win_pos);
	}
}


void Client::reconfigure(Client *c)
{	
	#ifdef DEBUG
		printf("started Client::reconfigure(Client *c)\n");
	#endif

	createTitleImage(c);
	createFocusImage(c);
	createButtonImage(c);	
	createPositionImage(c);
		
	do_gradient(c);
	redraw(c);
		
	#ifdef DEBUG
		printf("finished Client::reconfigure(Client *c)\n");
	#endif
}

void Client::send_config(Client *c)
{
	#ifdef DEBUG
		printf("started Client::send_config(Client *c)\n");
	#endif

	XConfigureEvent ce;
	XEvent foo;
	
	ce.type = ConfigureNotify;
	ce.event = c->window;
	ce.window = c->window;
	ce.x = c->x;
	ce.y = c->y;
	ce.width = c->width;
	ce.height = c->height;
	ce.border_width = 0;
	ce.above = None;
	ce.override_redirect = 0;
	
	if(! XCheckTypedWindowEvent(wm->getDisplay(), c->window, DestroyNotify, &foo))
	{
		reconfigure(c);
	}
	
	XSendEvent(wm->getDisplay(), c->window, False, StructureNotifyMask, (XEvent *)&ce);
	
	#ifdef DEBUG
		printf("finished Client::send_config(Client *c)\n");
	#endif	
}

void Client::move(Client *c)
{
	if (c) {	
		if(!c->shaded) {
			 XMapRaised(wm->getDisplay(), c->win_pos);
		}
		
		drag(c);
 		
		if(!c->shaded) XUnmapWindow(wm->getDisplay(), c->win_pos);	
							
		XMoveWindow(wm->getDisplay(), c->parent, c->x, c->y-theight(c));
		send_config(c);
		
	}
}

void Client::drag(Client *c)
{
	XEvent ev;
	int x1=c->x;
	int y1=c->y;
	int old_cx = c->x;
	int old_cy = c->y;

	wm->grabPointer(wm->getRootWindow(), ButtonPressMask | ButtonReleaseMask | PointerMotionMask, wm->move_curs);
	
	get_mouse_position(&x1, &y1);

	draw_outline(c);
	
	for (;;) {
		XMaskEvent(wm->getDisplay(), ButtonPressMask | ButtonReleaseMask | PointerMotionMask, &ev);
		switch (ev.type) {
			case MotionNotify:
				draw_outline(c); /* clear */
				c->x = old_cx + (ev.xmotion.x - x1);
				c->y = old_cy + (ev.xmotion.y - y1);
				draw_outline(c);
			break;
				
			case ButtonRelease:
				draw_outline(c); /* clear */
				wm->ungrabPointer();
				return;
		}
	}
}

void Client::resize(Client *c)
{
	if (c) {
		sweep(c);
		
		if (c->maxsized) {
			c->maxsized = False;
		}
		
		XMoveResizeWindow(wm->getDisplay(), c->parent, c->x, c->y - theight(c), c->width, c->height + theight(c));			
		XMoveResizeWindow(wm->getDisplay(), c->title, 0, 0, c->width, theight(c)); 		
		XMoveWindow(wm->getDisplay(), c->maximize.window, c->width - theight(c), c->maximize.y);
		XMoveWindow(wm->getDisplay(), c->minimize.window, c->width - theight(c) - 18, c->minimize.y);
		XMoveWindow(wm->getDisplay(), c->shadeButton.window, c->width - theight(c) - 36, c->shadeButton.y);
		XMoveResizeWindow(wm->getDisplay(), c->window, 0, theight(c), c->width, c->height);	
		
		send_config(c);
	}
}

void Client::recalculate_sweep(Client *c, int x1, int y1, int x2, int y2)
{
	int basex=0;
	int basey=0;

	c->width = abs(x1 - x2) - DEFAULT_BORDER_SIZE;
	c->height = abs(y1 - y2) - DEFAULT_BORDER_SIZE;

	if (c->size->flags & PResizeInc) {
		basex = (c->size->flags & PBaseSize) ? c->size->base_width :
			(c->size->flags & PMinSize) ? c->size->min_width : c->x;
		basey = (c->size->flags & PBaseSize) ? c->size->base_height :
			(c->size->flags & PMinSize) ? c->size->min_height : c->y;
		c->width -= (c->width - basex) % c->size->width_inc;
		c->height -= (c->height - basey) % c->size->height_inc;
	}
	
	if (c->size->flags & PMinSize) {
		if (c->width < c->size->min_width) c->width = c->size->min_width;
		if (c->height < c->size->min_height) c->height = c->size->min_height;
	}
	
	if (c->size->flags & PMaxSize) {
		if (c->width > c->size->max_width) c->width = c->size->max_width;
		if (c->height > c->size->max_height) c->height = c->size->max_height;
	}
	
	c->x = (x1 <= x2) ? x1 : x1 - c->width;
	c->y = (y1 <= y2) ? y1 : y1 - c->height;
}

void Client::sweep(Client *c)
{
	XEvent ev;
	int old_cx = c->x;
	int old_cy = c->y;
	char buffer[10];

	if(!c->shaded) XMapRaised(wm->getDisplay(), c->win_pos); 

	wm->grabPointer(wm->getRootWindow(), ButtonPressMask | ButtonReleaseMask | PointerMotionMask, wm->resize_curs);
	
	draw_outline(c);

	warpPointer(c->window, c->width, c->height);
	
	for (;;) {
		XMaskEvent(wm->getDisplay(), ButtonPressMask | ButtonReleaseMask | PointerMotionMask, &ev);
		switch (ev.type) {
			case MotionNotify:
				draw_outline(c); 
				recalculate_sweep(c, old_cx, old_cy, ev.xmotion.x, ev.xmotion.y);				
				draw_outline(c);
			break;
				
			case ButtonRelease:
				draw_outline(c);
				wm->ungrabPointer();
				if(!c->shaded) XUnmapWindow(wm->getDisplay(), c->win_pos);
				return;
		}
	}
}

void Client::draw_outline(Client *c)
{
	char buf[40];
	int text_width=0;
	int text_x=8;
	int text_y=15;
	
	if(c->shaded)
	{
		XDrawRectangle(wm->getDisplay(), wm->getRootWindow(), wm->invert_gc,
			c->x, 							/* x */
			c->y - theight(c) + DEFAULT_BORDER_SIZE/2,		/* y */
			c->width,						/* width */
			theight(c));						/* height */			
	
	} else {
		
		XDrawRectangle(wm->getDisplay(), wm->getRootWindow(), wm->invert_gc,
		c->x + DEFAULT_BORDER_SIZE/2,                   	/* x */
		c->y - theight(c) + DEFAULT_BORDER_SIZE/2,		/* y */
		c->width + DEFAULT_BORDER_SIZE, 			/* width */
		c->height + theight(c) + DEFAULT_BORDER_SIZE);	/* height */

		text_width=XTextWidth(wm->font, c->client_position, strlen(c->client_position));

		getClientPosition(c);

		XClearWindow(wm->getDisplay(), c->win_pos);
		
		XMoveResizeWindow(wm->getDisplay(), c->win_pos,  
			c->position_window_width,
			c->position_window_height,				 
			text_width-25,	 
			wm->iheight+2					 
			); 
					
		XDrawString(wm->getDisplay(), c->win_pos, wm->invert_gc,
			text_x,
			text_y,
			c->client_position, strlen(c->client_position));			
	}
}

void Client::getClientPosition(Client *c)
{
	int text_width=0;
	
	sprintf(c->client_position, "%dx%d+%d+%d", c->width, c->height+theight(c), c->x, c->y);
	text_width=XTextWidth(wm->font, c->client_position, strlen(c->client_position));
		
	c->position_window_width=wm->getXRes()/2 - (text_width/2)+15;
	c->position_window_height=wm->getYRes()/2;
}

void Client::maximumresize(Client *c)
{
	if (c) {

		if (c->maxsized)
		{
			c->x = c->oumx;
			c->y = c->oumy;
			c->width = c->oumwidth;
			c->height = c->oumheight;
			c->maxsized = False;
		} else {
			c->oumx = c->x;
			c->oumy = c->y;
			c->oumwidth = c->width;
			c->oumheight = c->height;

			c->x = 1;
			if (wm->theme->getToolbarLocationTop())
			{
				c->y = 28 + theight(c);
 	  		} else {
				c->y = 1 + theight(c);
 			}

			c->width = wm->getXRes() - 4;
			c->height = wm->getYRes() - 31 - theight(c);
			c->maxsized = True;
		}

		XMoveResizeWindow(wm->getDisplay(), c->parent, c->x, c->y - theight(c), c->width, c->height + theight(c));
		XMoveResizeWindow(wm->getDisplay(), c->title, 0, 0, c->width, theight(c)); 
		XMoveWindow(wm->getDisplay(), c->maximize.window, c->width - theight(c), c->maximize.y);
		XMoveWindow(wm->getDisplay(), c->minimize.window, c->width - theight(c) - 18, c->minimize.y);
		XMoveWindow(wm->getDisplay(), c->shadeButton.window, c->width - theight(c) - 36, c->shadeButton.y);
		XMoveResizeWindow(wm->getDisplay(), c->window, 0, theight(c), c->width, c->height);
			
		send_config(c);
	
	}
}

void Client::hide(Client *c)
{
	c->ignore_unmap++;
		
	XUnmapSubwindows(wm->getDisplay(), c->parent);
	XUnmapWindow(wm->getDisplay(), c->parent);
	set_wm_state(c, IconicState);
}

void Client::unhide(Client *c)
{
	XMapSubwindows(wm->getDisplay(), c->parent);
	XMapRaised(wm->getDisplay(), c->parent);
	set_wm_state(c, NormalState);
}

// This repaints the titlebar and buttons on it, It also paints the window title.
void Client::redraw(Client *c)
{
	XEvent foo;
	
	// Make sure the window still exists.
	if(! XCheckTypedWindowEvent(wm->getDisplay(), c->window, DestroyNotify, &foo))
	{	
		XClearWindow(wm->getDisplay(), c->window);
		
		do_gradient(c);
			
		// Close Button		
		XDrawLine(wm->getDisplay(), c->close.window, wm->border_gc, 2,2,c->close.width - 3,c->close.height-3);
		XDrawLine(wm->getDisplay(), c->close.window, wm->border_gc, 2,c->close.height - 3,c->close.width-3,2);

		// Maximize Button
		XDrawRectangle(wm->getDisplay(), c->maximize.window,
			wm->border_gc, 2, 2, c->maximize.width - 5, c->maximize.height - 5);
  		XDrawLine(wm->getDisplay(),  c->maximize.window,
	  		wm->border_gc, 2, 3, c->maximize.width - 3, 3);

		// Minimize Button
		XDrawRectangle(wm->getDisplay(), c->minimize.window,
				wm->border_gc, 2, c->minimize.height - 5, c->minimize.width - 5, 2);
				
		draw_shade_button(c);
				
		if (c->name) 
		{
			XDrawString(wm->getDisplay(), c->title, wm->string_gc, c->default_client_name_x, 
				c->default_client_name_y, c->name, strlen(c->name));		
		}
	}
}

void Client::draw_shade_button(Client *c)
{
	XClearWindow(wm->getDisplay(), c->shadeButton.window);	

	if (c->shaded)
	{		
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				3, 6, 12, 6);
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				4, 7, 11, 7);
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				5, 8, 10, 8);
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				6, 9, 9, 9);
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				7, 10, 8, 10);	
	  	
				
	} else {

		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				7, 6, 8, 6);
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				6, 7, 9, 7);
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				5, 8, 10, 8);
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				4, 9, 11, 9);
		XDrawLine(wm->getDisplay(),  c->shadeButton.window,wm->border_gc, 
				3, 10, 12, 10);

	}
}

void Client::change_gravity(Client *c, int multiplier)
{
	int dy = 0;
	int gravity = (c->size->flags & PWinGravity) ?
		c->size->win_gravity : NorthWestGravity;

	switch (gravity) {
		case NorthWestGravity:
		case NorthEastGravity:
		case NorthGravity: dy = theight(c); break;
		case CenterGravity: dy = theight(c)/2; break;		
	}

	c->y += multiplier * dy;
}

#ifdef SHAPE
void Client::set_shape(Client *c)
{
//	int n, order;
//	XRectangle *rect;
//
//	rect = XShapeGetRectangles(wm->getDisplay(), c->window, ShapeBounding, &n, &order);
//	if (n > 1) XShapeCombineShape(wm->getDisplay(),
//		c->parent, ShapeBounding, 0, theight(c),
//		c->window, ShapeBounding, ShapeSet);
//	XFree(rect);
}
#endif

void Client::get_mouse_position(int *x, int *y)
{
	Window dw1=None;
	Window dw2=None;
	int t1=0;
	int t2=0;
	unsigned int t3=0;

	XQueryPointer(wm->getDisplay(), wm->getRootWindow(), &dw1, &dw2, x, y, &t1, &t2, &t3);
}

void Client::shade(Client *c)
{
	if (c->shaded)
	{
	   	XResizeWindow(wm->getDisplay(), c->parent, c->width, c->height + theight(c));
	   	c->shaded=False;
	} else {
	   	XResizeWindow(wm->getDisplay(), c->parent, c->width, theight(c) - 1);
	  	c->shaded=True;
	}
}

void Client::closeButtonPressed(Client *c)
{
	XSetWindowBorderWidth(wm->getDisplay(), c->close.window, 2);
	
	XEvent ev;
	wm->grabPointer(c->parent, ButtonReleaseMask, wm->left_arrow);
	
	for (;;) {
		XMaskEvent(wm->getDisplay(), ButtonReleaseMask, &ev);
		switch (ev.type) {
			case ButtonRelease:
			{
				XSetWindowBorderWidth(wm->getDisplay(), c->close.window, 1);
				send_wm_delete(c);
				wm->ungrabPointer();
				return;
			}
		}
	}
}

void Client::shadeButtonPressed(Client *c)
{
	XSetWindowBorderWidth(wm->getDisplay(), c->shadeButton.window, 2);

	XEvent ev;
	wm->grabPointer(c->parent, ButtonReleaseMask, wm->left_arrow);
	
	for (;;) {
		XMaskEvent(wm->getDisplay(), ButtonReleaseMask, &ev);
		switch (ev.type) {
			case ButtonRelease:
			{
				XSetWindowBorderWidth(wm->getDisplay(), c->shadeButton.window, 1);
				XRaiseWindow(wm->getDisplay(), c->parent);
				shade(c);
				wm->ungrabPointer();
				return;
			}
		}
	}
}

void Client::minimizeButtonPressed(Client *c)
{
	XSetWindowBorderWidth(wm->getDisplay(), c->minimize.window, 2);
	
	XEvent ev;
	wm->grabPointer(c->parent, ButtonReleaseMask, wm->left_arrow);
	
	for (;;) {
		XMaskEvent(wm->getDisplay(), ButtonReleaseMask, &ev);
		switch (ev.type) {
			case ButtonRelease:
			{
				XSetWindowBorderWidth(wm->getDisplay(), c->minimize.window, 1);
				XLowerWindow(wm->getDisplay(), c->parent); 	
				wm->ungrabPointer();
				return;
			}
		}
	}
}

void Client::maximizeButtonPressed(Client *c)
{
	XSetWindowBorderWidth(wm->getDisplay(), c->maximize.window, 2);
	
	XEvent ev;
	wm->grabPointer(c->parent, ButtonReleaseMask, wm->left_arrow);
	
	for (;;) {
		XMaskEvent(wm->getDisplay(), ButtonReleaseMask, &ev);
		switch (ev.type) {
			case ButtonRelease:
			{
				XSetWindowBorderWidth(wm->getDisplay(), c->maximize.window, 1);
				
				switch(ev.xbutton.button)
				{
				
					case 1: {
						if(c->shaded)
						{
							shade(c);
							XRaiseWindow(wm->getDisplay(), c->parent);
							resize(c);
						} else {
							XRaiseWindow(wm->getDisplay(), c->parent);
							resize(c);
						}
				
					wm->ungrabPointer();
					return;
					}
				
					case 2:
					{
					
					
					wm->ungrabPointer();
					return;
					}
			
					case 3:
					{
						if(c->shaded)
						{
							shade(c);
							XRaiseWindow(wm->getDisplay(), c->parent);
						
							// Make sure it doesn't keep resizing if its already max resized
							if (! c->maxsized) maximumresize(c);
						
						} else {
							XRaiseWindow(wm->getDisplay(), c->parent);
							maximumresize(c);
						}
					
					wm->ungrabPointer();
					return;
					}
				}
			}
		}
	}
}

void Client::send_wm_delete(Client *c)
{
	#ifdef DEBUG
		printf("started Client::send_wm_delete(Client *c)\n");
	#endif

	if (c) {
		if (c->proto & Pdelete) {
			send_xmessage(c->window, wm->getWMProtocolsAtom(), wm->getWMDeleteAtom(), CurrentTime);
		} else {
			XKillClient(wm->getDisplay(), c->window);
		}
	}
	
	#ifdef DEBUG
		printf("finished Client::send_wm_delete(Client *c)\n");
	#endif
}

int Client::send_xmessage(Window w, Atom a, long x, long y)
{
	#ifdef DEBUG
		printf("started int Client::send_xmessage(Window w, Atom a, long x)\n");
	#endif

	XEvent ev;
	long mask;

	memset(&ev, '\0', sizeof ev);
	ev.type = ClientMessage;
	ev.xclient.window = w;
	ev.xclient.message_type = a;
	ev.xclient.format = 32;
	ev.xclient.data.l[0] = x;
	ev.xclient.data.l[1] = y;
	mask = (w == wm->getRootWindow()) ? SubstructureRedirectMask : 0L;
	
	#ifdef DEBUG
		printf("finished int Client::send_xmessage(Window w, Atom a, long x)\n");
	#endif
	
	return XSendEvent(wm->getDisplay(), w, False, mask, &ev);
}

void Client::set_wm_state(Client *c, int state)
{
	#ifdef DEBUG
		printf("started Client::set_wm_state(Client *c, int state)\n");
	#endif

	long data[2];

	data[0] = (long) state;
	data[1] = (long) None; /* icon window */

	c->state = state;
	
	XChangeProperty(wm->getDisplay(), c->window, wm->getWMStateAtom(), wm->getWMStateAtom(),
		32, PropModeReplace, (unsigned char *)data, 2);

	#ifdef DEBUG
		printf("finished Client::set_wm_state(Client *c, int state)\n");
	#endif
}

int Client::get_wm_state(Client *c)
{
	#ifdef DEBUG
		printf("started Client::get_wm_state(Client *c)\n");
	#endif

	Atom real_type;
	int real_format, state;
	unsigned long n, extra;
	unsigned char *data;

	if ((XGetWindowProperty(wm->getDisplay(), c->window, wm->getWMStateAtom(), 0L, 2L, False,
				AnyPropertyType, &real_type, &real_format, &n,
				&extra, &data) == Success) && n) {
		state = *(int *)data;
		XFree(data);
		return state;
	} else return WithdrawnState;
	
	#ifdef DEBUG
		printf("finished Client::get_wm_state(Client *c)\n");
	#endif	
}

void Client::handle_motion_event(XMotionEvent *e)
{
}

void Client::handle_focus_in_event(XFocusChangeEvent *e)
{
	Client *c = find_client(e->window);	

	if(c)
	{	
		c->has_focus=True;
		do_gradient(c);
		redraw(c);
	} 	
}

void Client::handle_focus_out_event(XFocusChangeEvent *e)
{
	Client *c = find_client(e->window);	
	
	if (c)
	{
		c->has_focus=False;
		do_gradient(c);
		redraw(c);
	}
}

void Client::handle_enter_event(XCrossingEvent *e)
{
	Client *c = find_client(e->window);	

	if(c)
	{
		XSetInputFocus(wm->getDisplay(), c->window, RevertToPointerRoot, CurrentTime);
		
		// For some strange reason if I use the next couple lines of code and I enter an xterm window
		// with the pointer. Xterm either crashes or it gets a DestroyNotify event and dies. This only
		// happens with Xterm as far as I know.
		
		//if (c->proto & Ptakefocus) {
		//	send_xmessage(c->window, wm->getWMProtocolsAtom(), wm->getWMFocusAtom(), CurrentTime);
		// }

	} 
}

void Client::handle_button_press_event(XButtonEvent *e)
{
	Client *c = find_client(e->window);

	if (c) {
		
		switch (e->button) {
			case Button1:
			{
				if(c->close.window == e->window)
				{
					closeButtonPressed(c);
										
				} else if (c->maximize.window == e->window)
				{
					maximizeButtonPressed(c);
					
				} else if (c->minimize.window == e->window)
				{
					minimizeButtonPressed(c);
					
				} else if (c->shadeButton.window == e->window) {
					shadeButtonPressed(c);
				
				} else if (c->title == e->window){
					XRaiseWindow(wm->getDisplay(), c->parent);
					redraw(c);
					move(c);
				} 
			break;
			}

			case Button2:
			{	
				 	if (c->title == e->window) {

						// Middle click also allows you to shade the window.
						if(c->shaded)
						{
							XRaiseWindow(wm->getDisplay(), c->parent);
							shade(c);
							redraw(c);
						} else {
							XRaiseWindow(wm->getDisplay(), c->parent);
							shade(c);
							redraw(c);
						}
					}	

			break;
			}
			case Button3:
			{
				if (c->maximize.window == e->window) maximizeButtonPressed(c);
			break;
			}
		}
	}
}

void Client::handle_button_release_event(XButtonEvent *e)
{
/*	Client *c = find_client(e->window);

	if (c) {
		switch (e->button) {
			case Button1:
			{
			break;
			}

			case Button2:
			{	
			break;
			}
			
			case Button3:
			{
			break;
			}
		}
	}
*/
}

void Client::handle_configure_request(XConfigureRequestEvent *e)
{
	#ifdef DEBUG
		printf("started Client::handle_configure_request(XConfigureRequestEvent *e)\n");
	#endif

	Client *c = find_client(e->window);
	XWindowChanges wc;

	if (c) {
		change_gravity(c, -1);
		
		if (e->value_mask & CWX) c->x = e->x;
		if (e->value_mask & CWY) c->y = e->y;
		if (e->value_mask & CWWidth) c->width = e->width;
		if (e->value_mask & CWHeight) c->height = e->height;
		
		change_gravity(c, 1);

		if (c->width != wc.width && c->height != wc.height)
		{
			if (c->has_titlebar) XMoveResizeWindow(wm->getDisplay(), c->parent, c->x, c->y - theight(c), c->width, c->height + theight(c));
			
			XMoveResizeWindow(wm->getDisplay(), c->title, -1, -1, c->width, theight(c));
			XMoveWindow(wm->getDisplay(), c->maximize.window, c->width - theight(c), c->maximize.y);
			XMoveWindow(wm->getDisplay(), c->minimize.window, c->width - theight(c) - 18, c->minimize.y);
			XMoveWindow(wm->getDisplay(), c->shadeButton.window, c->width - theight(c) - 36, c->shadeButton.y);
			
			if (c->has_titlebar) XMoveResizeWindow(wm->getDisplay(), c->window, 0, theight(c), c->width, c->height);
		}
		
		wc.x = c->x;
		wc.y = c->y - theight(c);
		wc.width = c->width;
		
		if (c->has_titlebar) wc.height = c->height + theight(c);
		else wc.height = c->height;
		
		wc.border_width = DEFAULT_BORDER_SIZE;
		wc.sibling = e->above;
		wc.stack_mode = e->detail;
		
		XConfigureWindow(wm->getDisplay(), c->parent, e->value_mask, &wc);

		send_config(c);
	}

	wc.x = c ? 0 : e->x;
	wc.y = c ? 0 : e->y;
	wc.width = e->width;
	wc.height = e->height;
	wc.sibling = e->above;
	wc.stack_mode = e->detail;

	XConfigureWindow(wm->getDisplay(), e->window, e->value_mask, &wc);
	
	#ifdef DEBUG
		printf("finished Client::handle_configure_request(XConfigureRequestEvent *e)\n");
	#endif	
}

void Client::handle_map_request(XMapRequestEvent *e)
{
	#ifdef DEBUG
		printf("started Client::handle_map_request(XMapRequestEvent *e)\n");
	#endif

	Client *c = find_client(e->window);

	if (c) {
		XMapWindow(wm->getDisplay(), c->parent);
		XMapWindow(wm->getDisplay(), c->window);
		
		set_wm_state(c, NormalState);
		
	} else
		make_new_client(e->window);

	#ifdef DEBUG
		printf("finished Client::handle_map_request(XMapRequestEvent *e)\n");
	#endif	
}

void Client::handle_unmap_event(XUnmapEvent *e)
{
	#ifdef DEBUG
		printf("started Client::handle_unmap_event(XUnmapEvent *e)\n");
	#endif	

	Client *c = find_client(e->window);

	if (c) {
		if (c->ignore_unmap) c->ignore_unmap--;
		else {
			XUnmapWindow(wm->getDisplay(), c->parent);
			XUnmapWindow(wm->getDisplay(), c->window);
		}
	} 
	
	#ifdef DEBUG
		printf("finished Client::handle_unmap_event(XUnmapEvent *e)\n");
	#endif	
}

void Client::handle_client_message(XClientMessageEvent *e)
{
	#ifdef DEBUG
		printf("started Client::handle_client_message(XClientMessageEvent *e)\n");
	#endif			

	Client *c = find_client(e->window);
	
	if (c)
	{
		if(e->message_type == wm->getWMChangeStateAtom())
		{
			if(e->format == 32 && e->data.l[0] == IconicState) hide(c);
		}
	}
		
	#ifdef DEBUG
		printf("finished Client::handle_client_message(XClientMessageEvent *e)\n");
	#endif			
}

void Client::handle_colormap_change(XColormapEvent *e)
{
	#ifdef DEBUG
		printf("started Client::handle_colormap_change(XColormapEvent *e)\n");
	#endif	

	Client *c =  find_client(e->window);

	if (c && e->c_new) {
		c->cmap = e->colormap;
		installColorMap(c->cmap);
	}	

	#ifdef DEBUG
		printf("finished Client::handle_colormap_change(XColormapEvent *e)\n");
	#endif	
}

void Client::handle_property_change(XPropertyEvent *e)
{
	#ifdef DEBUG
		printf("started Client::handle_property_change(XPropertyEvent *e)\n");
	#endif	

	Client *c = find_client(e->window);
	long dummy;

	if (c) {
		switch (e->atom) {
			case XA_WM_NAME:
			{
				if (c->name) XFree(c->name);
				XFetchName(wm->getDisplay(), c->window, &c->name);
				XClearWindow(wm->getDisplay(), c->parent);
				redraw(c);
			}
			break;
			
			case XA_WM_NORMAL_HINTS:
			{
				XGetWMNormalHints(wm->getDisplay(), c->window, c->size, &dummy);				
			}
			break;
						
			case XA_WM_ICON_NAME:
			break;
			
			case XA_WM_TRANSIENT_FOR:
				getTransients(c);
			break;
		}
	}

	#ifdef DEBUG
		printf("finished Client::handle_property_change(XPropertyEvent *e)\n");
	#endif	
}


void Client::handle_expose_event(XExposeEvent *e)
{
	Client *c = find_client(e->window);

	if (c && e->count == 0) { redraw(c); }

	//#ifdef DEBUG
		//printf("finished Client::handle_expose_event(XExposeEvent *e)\n");
	//#endif	
}

void Client::handle_reparent_event(XReparentEvent *e)
{
	#ifdef DEBUG
		printf("started Client::handle_reparent_event(XReparentEvent *e)\n");
	#endif	

	Client *c = find_client(e->window);
	Client *dest = find_client(e->parent);

	if (c && !dest) { 
			
		remove_client(c);
	} 
	
	#ifdef DEBUG
		printf("finished Client::handle_reparent_event(XReparentEvent *e)\n");
	#endif	
}

void Client::handle_destroy_notify(XDestroyWindowEvent *e)
{
	Client *c = find_client(e->window);
	
	if (c)
	{
		remove_client(c);	
	}
	
	#ifdef DEBUG
		printf("finished Client::handle_destroy_notify(XDestroyWindowEvent *e)\n");
	#endif			
}

#ifdef SHAPE
void Client::handle_shape_change(XShapeEvent *e)
{
//	Client *c = find_client(e->window);
//
//	if (c) set_shape(c);
//
//	#ifdef DEBUG
//		printf("finished Client::handle_shape_change(XShapeEvent *e)\n");
//	#endif		
}
#endif

void Client::warpPointer(Window w, int x, int y)
{
	XWarpPointer(wm->getDisplay(), None, w, 0, 0, 0, 0, x, y);
}

void Client::installColorMap(Colormap cmap)
{
	XInstallColormap(wm->getDisplay(), cmap);
}

void Client::isIconic(Client *c)
{
	#ifdef DEBUG
		printf("Started Client::isIconic(Client *c)\n");
	#endif		

	// Determine if this client is already in its iconic state upon creation
	// if so then we won't map it.
	if (get_wm_state(c) == IconicState) {
		
		if (c->has_titlebar) 
		{
			XUnmapSubwindows(wm->getDisplay(), c->parent);
			XUnmapWindow(wm->getDisplay(), c->parent);
		}
		
		XUnmapWindow(wm->getDisplay(), c->window);
	} else {
	
		mapClient(c);
	}

	#ifdef DEBUG
		printf("Finished Client::isIconic(Client *c)\n");
	#endif		
}

void Client::mapClient(Client *c)
{
	#ifdef DEBUG
		printf("Started Client::mapClient(Client *c)\n");
	#endif		
	
	if (c->has_titlebar)
	{		
		if(c->has_shade_button)    XMapWindow(wm->getDisplay(), c->shadeButton.window);
		if(c->has_close_button)    XMapWindow(wm->getDisplay(), c->close.window);
		if(c->has_minimize_button) XMapWindow(wm->getDisplay(), c->minimize.window);
		if(c->has_maximize_button) XMapWindow(wm->getDisplay(), c->maximize.window);
		
		XMapWindow(wm->getDisplay(), c->title);
	}

	XMapWindow(wm->getDisplay(), c->window);		
	XMapRaised(wm->getDisplay(), c->parent);
	
	set_wm_state(c, NormalState);		
	
	#ifdef DEBUG
		printf("Finished Client::mapClient(Client *c)\n");
	#endif		
}

